/*
John (Jake) Ware and Jarvis Schultz
Northwestern University
ME333
Mobile Robot
3/15/2010
*/

//Includes
import processing.serial.*;
import controlP5.*;

//Objects
Serial myPort;        // The serial port
ControlP5 controlP5;
Textfield TextField1;

//Global Variables
int MAX_V = 14;      //maximum velocity in [in/s]
int MIN_V = 4;       //minimum velocity in [in/s]
int WIDTH = 2800;    //terrain width in hundreths of inches
int HEIGHT = 3200;   //terrain height in hundreths of inches
int origin_x = 0;    //origin location in the processing window (relative to the center)
int origin_y = 0;
int count = 0;       //count of requested locations (number of times you have clicked to create a new waypoint)
int mouse_x = 0;     //location of the most recent mouse click
int mouse_y = 0;
int mouse_x_last = 0;    //location of previous click
int mouse_y_last = 0;
boolean newdata = false;    //flag for new destination
String output;              //output string sent to robot
boolean p_control = true;   //flag for position control or velocity control

//********************************************Setup**************************************
void setup () {
  frameRate(100);
  size(500, 600);   
  origin_x = width/2;
  origin_y = height/2; 
  mouse_x = origin_x;
  mouse_y = origin_y;
  mouse_x_last = origin_x;
  mouse_y_last = origin_y;
  println("X Origin is at " + origin_x);
  println("Y Origin is at " + origin_y);
  println(Serial.list());
  myPort = new Serial(this, Serial.list()[0], 115200);
  myPort.bufferUntil('\n');
  background(120);
  rect(50,50,height-100,width-100);    //window in screen
  PFont font;
  font = loadFont("Calibri-24.vlw");
  textFont(font);
  
  //ControlP5 Setup
  //The ControlP5 website has all relevant details on these objects
  controlP5 = new ControlP5(this);
  
  //Create buttons for sending data to start or stop
  controlP5.addButton("Reset",1,5,10,80,19);
  controlP5.addButton("Stop",1,90,10,80,19);
  controlP5.addButton("Home",1,175,10,80,19);
  controlP5.addButton("Mode",1,515,10,80,19);
  TextField1 = controlP5.addTextfield("Set_Speed",260,9,100,20);
}

//********************************************Draw**************************************
void draw () {
  //draw the robot path from the last point to this point or the velocity vector
  stroke(255,0,0);
  if(p_control) {
    line(mouse_x_last,mouse_y_last,mouse_x,mouse_y);
  } else {
    line(origin_x,origin_y,mouse_x,mouse_y);
  }
  
  //draw indicator box for state
  stroke(0);
  if(p_control) {
    fill(0);
  } else {
    fill(255);
  }
  rect(460,10,50,18);
  
  //draw legend
  stroke(255);
  fill(0);
  text("Position Control",80,575);
  rect(50,557,20,20);
  stroke(0);
  fill(255);
  text("Velocity Control",300,575);
  rect(270,557,20,20);
  
  //if there is new position data then format this into a string and send it
  if(newdata == true) {
    if(p_control) {
      FormatDataP();
    } else {
      FormatDataV();
    }
    SendData();
    newdata = false;
  }
}

//********************************************Functions**************************************
//this function is not used in this version, but is a skeleton for receiving data from the robot
void serialEvent (Serial myPort) {
  String inString = myPort.readStringUntil('\n');
  if (inString != null) {
    inString = trim(inString);  // trim off any whitespace:
    //do stuff here
  }
}

//this function gets called when the mouse is clicked and draws a waypoint at that location
void mousePressed() {
  if(mouseX > 50 && mouseX < 550 && mouseY > 50 && mouseY < 550) {
    mouse_x_last = mouse_x;
    mouse_y_last = mouse_y;
    mouse_x = mouseX;
    mouse_y = mouseY;  
    stroke(0);
    if(p_control) {
      fill(0);
      stroke(0);
      ellipse(mouse_x-4,mouse_y-4,8,8);
      text(++count,mouse_x,mouse_y-5);
    } else {
      fill(255);
      stroke(0);
      background(120);
      rect(50,50,height-100,width-100);
      fill(0);
      ellipse(mouse_x-4,mouse_y-4,8,8);
      text("V",mouse_x-5,mouse_y-5);
    }
    newdata = true;
  } 
}

//scale the coordinates of the clicked point to the surface the robot is on
void FormatDataP() {
  int diff_x = 0;
  int diff_y = 0;
  int scaled_diff_x = 0;
  int scaled_diff_y = 0;
  String output_x;
  String output_y;
  String theta_str;
  int theta = 0;
  
  diff_x = mouse_x - origin_x;
  diff_y = origin_y - mouse_y;
    
  //scale x
  scaled_diff_x = floor(map(diff_x, -250, 250, -WIDTH/2, WIDTH/2));
  //make x a string of the correct length
  output_x = Integer.toString(abs(scaled_diff_x));
  while(output_x.length() < 4) output_x = "0" + output_x;
  //add a 1 for positive or 0 for negative
  output_x = (scaled_diff_x >= 0) ? ("1" + output_x) : ("0" + output_x);
  
  //scale y 
  scaled_diff_y = floor(map(diff_y, -250, 250, -HEIGHT/2, HEIGHT/2));
  //make y a string of the correct length
  output_y = Integer.toString(abs(scaled_diff_y));
  while(output_y.length() < 4) output_y = "0" + output_y;
  //add a 1 for positive or 0 for negative
  output_y = (scaled_diff_y >= 0) ? ("1" + output_y) : ("0" + output_y);
    
  //println(scaled_diff_x + " " + scaled_diff_y);
    
  //create angle
  if((mouse_y_last - mouse_y) >= 0) {
    theta = floor(1000*atan2((float)(mouse_y_last - mouse_y),(float)(mouse_x - mouse_x_last)));
  } else {
    theta = floor(1000*(2*PI + atan2((float)(mouse_y_last - mouse_y),(float)(mouse_x - mouse_x_last))));
  }
  theta_str = Integer.toString(theta);
  while(theta_str.length() < 4) theta_str = "0" + theta_str;
  
  //create output string
//  output = "p"+output_x + output_y + theta_str;
  output = "p"+output_x + output_y + "0000";
}

//scale the coordinates of the point clicked into a velocity vector that the robot will follow
void FormatDataV() {
  float diff_x = 0;
  float diff_y = 0;
  int Vx = 0;
  int Vy = 0;
  String output_Vx;
  String output_Vy;
  String theta_str;
  float theta = 0.0;
  
  diff_x = abs(mouse_x - origin_x);
  diff_y = abs(origin_y - mouse_y);
  if((origin_y - mouse_y) >= 0) {
    theta = atan2((float)(origin_y - mouse_y),(float)(mouse_x - origin_x));
  } else {
    theta = 2*PI + atan2((float)(origin_y - mouse_y),(float)(mouse_x - origin_x));
  }
  
  //scale x
  Vx = (int) floor(100*map(diff_x, 0, 250, MIN_V, MAX_V)*cos(theta));
  //make x a string of the correct length
  output_Vx = Integer.toString(abs(Vx));
  while(output_Vx.length() < 4) output_Vx = "0" + output_Vx;
  //add a 1 for positive or 0 for negative
  output_Vx = (Vx >= 0) ? ("1" + output_Vx) : ("0" + output_Vx);
  
  //scale y 
  Vy = (int) floor(100*map(diff_y, 0, 250, MIN_V, MAX_V)*sin(theta));
  //make y a string of the correct length
  output_Vy = Integer.toString(abs(Vy));
  while(output_Vy.length() < 4) output_Vy = "0" + output_Vy;
  //add a 1 for positive or 0 for negative
  output_Vy = (Vy >= 0) ? ("1" + output_Vy) : ("0" + output_Vy);
    
  //println(Vx + " " + Vy + " " + theta);
  
  //create output string
  output = "v"+output_Vx + output_Vy + "0000";
}

//this function sends the output string to the robot
void SendData() {
  println(output);
  myPort.write(output);
}

//this function sends the proper string to move the origin to the robots current position
public void Reset(int value) {
  mouse_x_last = origin_x;
  mouse_y_last = origin_y;
  mouse_x = origin_x;
  mouse_y = origin_y;
  count = 0;
  fill(255);
  stroke(0);
  background(120);
  rect(50,50,height-100,width-100);
  
  println("Reseting Coordinate System to Current Position");
  output = "r00000000000000";
  SendData();
}

//this function sends the proper string to stop the robot
public void Stop(int value) {
  println("Stopping Robot");
  output = "q00000000000000";
  SendData();
}

//this function sends the proper string to make the robot return to its current origin location
public void Home(int value) {
  println("Sending Robot Back to Origin");
  output = "p00000000000000";
  SendData();
}

//this function changes the GUI mode of operation and cleans the screen up
public void Mode(int value) {
  mouse_x_last = origin_x;
  mouse_y_last = origin_y;
  mouse_x = origin_x;
  mouse_y = origin_y;
  fill(255);
  stroke(0);
  background(120);
  rect(50,50,height-100,width-100);
  if(p_control == true) {
    p_control = false;
    println("Mode: Velocity Control");
  } else {
    p_control = true;
    println("Mode: Position Control");
  }
}

//this function is called when something is entered into the textbox and sends the proper string to set the wheel speeds with the entered value
public void Set_Speed(String speed) {
  if(Integer.parseInt(speed) >= 20 && Integer.parseInt(speed) <= 900)
  {
    output = "s" + speed + "00000000000";
    SendData();
  }
}
